// Copyright 2023 The Pigweed Authors
//
// Licensed under the Apache License, Version 2.0 (the "License"); you may not
// use this file except in compliance with the License. You may obtain a copy of
// the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations under
// the License.

`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// SPI m Controller
//
//////////////////////////////////////////////////////////////////////////////////



module spi_m_core_ctl (
    input mclk_i,
    input rst_i,
    input r_w_en_i,
    input en_i,
    input mode_i,
    input dn_clr_i,  // Added to sync ADC polling logic with spi_s_core.v
    input [5:0] rwaddr_i,
    input [15:0] wdata_i,
    output reg dn_o,
    output reg poll_dn_o,
    output reg [39:0] rdata_o,
    output reg [23:0] vbus_o,
    output reg [23:0] vshunt_o,
    // spi interface
    input miso_i,
    input alert_n_i,
    output sclk_o,
    output mosi_o,
    output cs_o
);

  reg
      r_w_en_r,
      alert_n_r2,
      alert_n_r1,
      alert_n_r0,
      spi_poll_en_r,
      spi_core_en_r,
      spi_m_en_r,
      rst_i_r,
      wait_ctr_en_r;
  reg dn_r;
  reg [1:0] poll_addr_select;
  reg [2:0] wait_ctr_r;
  reg [2:0] poll_state;
  reg [5:0] poll_rwaddr_r;
  reg [5:0] rwaddr_r;
  reg [15:0] wdata_r;
  reg [39:0] spi_rdata_r;
  wire [39:0] spi_rdata_w;
  wire dn_w;

  localparam S0 = 0;
  localparam S1 = 1;
  localparam S2 = 2;
  localparam S3 = 3;
  localparam S4 = 4;
  localparam poll_r_w_en = 1;  // Continous read.
  localparam poll_addr0 = 6'b000100;  // 04h = V-shunt Register
  localparam poll_addr1 = 6'b000101;  // 05h = V-bus Register
  localparam poll_addr2 = 6'b001011;  // 0Bh = Diag Alert Reg. to clear Alert.

  spi_m_core spi_m_inst (
      .mclk_i(mclk_i),
      .rst_i(rst_i_r),
      .r_w_en_i(r_w_en_r),
      .en_i(spi_m_en_r),
      .rwaddr_i(rwaddr_r),
      .wdata_i(wdata_r),
      .dn_o(dn_w),
      .rdata_o(spi_rdata_w),
      // spi
      .miso_i(miso_i),
      .sclk_o(sclk_o),
      .mosi_o(mosi_o),
      .cs_o(cs_o)
  );

  always @(posedge mclk_i) begin
    rst_i_r <= rst_i;
  end

  always @(posedge mclk_i) begin
    if (rst_i_r) spi_m_en_r <= 0;
    else spi_m_en_r <= (en_i | spi_poll_en_r);
  end

  always @(posedge mclk_i) begin
    if (rst_i_r) begin
      dn_o <= 0;
      rdata_o <= 0;
    end else begin
      if (!mode_i) begin
        dn_o <= dn_w;
        if (dn_w) rdata_o <= spi_rdata_w;
      end
    end
  end

  always @(posedge mclk_i) begin
    if (rst_i_r) begin
      {alert_n_r0, alert_n_r1, alert_n_r2} <= 3'b111;
    end else begin
      {alert_n_r2, alert_n_r1, alert_n_r0} <= {alert_n_r1, alert_n_r0, alert_n_i};
    end
  end

  always @(posedge mclk_i) begin
    if (mode_i) begin  // if mode=0 R/W Mode. // if mode=1 Continous Mode
      rwaddr_r <= poll_rwaddr_r;
      wdata_r  <= 0;
      r_w_en_r <= poll_r_w_en;
    end else begin
      rwaddr_r <= rwaddr_i;
      wdata_r  <= wdata_i;
      r_w_en_r <= r_w_en_i;
    end
  end

  // Shunt reg.s Polling state machine

  always @(posedge mclk_i) begin
    if (rst_i_r) poll_state <= S0;
    else begin
      if (mode_i) begin
        case (poll_state)
          S0: begin
            if (alert_n_r2 | dn_clr_i) poll_state <= S0;
            else poll_state <= S1;
          end
          S1: begin
            poll_state <= S2;
          end
          S2: begin
            if (dn_w) poll_state <= S3;
            else poll_state <= S2;
          end
          S3: begin
            if (poll_addr_select>2) begin 		// Changed to ">2" to clear alert after reading the vbus and vshunt registers.
              poll_state <= S4;
            end else begin
              if (wait_ctr_r[2])  // (wait_ctr_r>=5) // Changed to one bit comparator.
                poll_state <= S1;
              else poll_state <= S3;
            end
          end
          S4: begin
            if (dn_clr_i)
              poll_state <= S0; // alert_n_r2  // Changed to pause countinous reads updates causing this module to go out of sync wrt to spi_s_core.v
            else begin
              poll_state <= S4;
            end
          end
          default: poll_state <= S0;
        endcase
      end else poll_state <= S0;
    end
  end

  always @(posedge mclk_i) begin
    if (rst_i_r) begin
      spi_poll_en_r <= 0;
      wait_ctr_en_r <= 0;
      poll_dn_o <= 0;
    end else begin
      case (poll_state)
        S0: begin
          poll_dn_o <= 0;
          wait_ctr_en_r <= 0;
          spi_poll_en_r <= 0;
        end
        S1: begin
          poll_dn_o <= 0;
          spi_poll_en_r <= 1;
          wait_ctr_en_r <= 0;
        end
        S2: begin
          spi_poll_en_r <= 0;
          poll_dn_o <= 0;
          wait_ctr_en_r <= 0;
        end
        S3: begin  // min wait time b/w 2 transaction
          poll_dn_o <= (poll_addr_select > 2) ? 1 : 0;
          spi_poll_en_r <= 0;
          wait_ctr_en_r <= 1;
        end
        S4: begin
          spi_poll_en_r <= 0;
          poll_dn_o <= 0;
          wait_ctr_en_r <= 0;
        end
      endcase
    end
  end

  always @(posedge mclk_i) begin
    if (rst_i_r | poll_dn_o) poll_addr_select <= 0;
    if (spi_m_en_r & mode_i) poll_addr_select <= poll_addr_select + 1;
  end

  always @(posedge mclk_i) begin
    if (poll_addr_select == 0) poll_rwaddr_r <= poll_addr0;
    else if (poll_addr_select == 1) poll_rwaddr_r <= poll_addr1;
    else poll_rwaddr_r <= poll_addr2;
  end

  always @(posedge mclk_i) begin
    if (dn_w & mode_i) begin
      if (poll_addr_select == 2'b01) vshunt_o <= spi_rdata_w;
      else if (poll_addr_select == 2'b10) vbus_o <= spi_rdata_w;
    end
  end

  always @(posedge mclk_i) begin
    if (wait_ctr_en_r) wait_ctr_r <= wait_ctr_r + 1;
    else wait_ctr_r <= 0;
  end

endmodule
